# -*- coding: utf-8 -*-
"""

   `Stats Tools`
    Drop-in functions when neither numpy or statistics can be imported
--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--
    The MIT License (MIT) (c) 2016 Jared Lunde

    Credit: http://stackoverflow.com/a/27758326
    Credit: http://stackoverflow.com/a/24101534

"""


class np(object):
    pass


def mean(data):
    """Return the sample arithmetic mean of data."""
    #: http://stackoverflow.com/a/27758326
    n = len(data)
    if n < 1:
        raise ValueError('mean requires at least one data point')
    return sum(data)/n  # in Python 2 use sum(data)/float(n)


def _ss(data):
    """Return sum of square deviations of sequence data."""
    #: http://stackoverflow.com/a/27758326
    c = mean(data)
    ss = sum((x-c)**2 for x in data)
    return ss


def pstdev(data):
    """Calculates the population standard deviation."""
    #: http://stackoverflow.com/a/27758326
    n = len(data)
    if n < 2:
        raise ValueError('variance requires at least two data points')
    ss = _ss(data)
    pvar = ss/n  # the population variance
    return pvar**0.5


def median(lst):
    """ Calcuates the median value in a @lst """
    #: http://stackoverflow.com/a/24101534
    sortedLst = sorted(lst)
    lstLen = len(lst)
    index = (lstLen - 1) // 2
    if (lstLen % 2):
        return sortedLst[index]
    else:
        return (sortedLst[index] + sortedLst[index + 1])/2.0



# -*- coding: utf-8 -*-
"""

  `Vital HTTP Tools`
--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--
   The MIT License (MIT) (c) 2015 Marcel Hellkamp

"""
import time
import base64
import email
import datetime

from vital.tools.encoding import uniorbytes


__all__ = (
  "http_date",
  "parse_date",
  "parse_auth",
  "parse_range_header"
)


def http_date(value):
    """ Formats the @value in required HTTP style

        @value: :class:datetime.datetime, #int, #float, #str time-like object

        -> #str HTTP-style formatted date

        (c)2014, Marcel Hellkamp
    """
    if isinstance(value, datetime.datetime):
        value = value.utctimetuple()
    elif isinstance(value, (int, float)):
        value = time.gmtime(value)
    if not isinstance(value, str):
        value = time.strftime("%a, %d %b %Y %H:%M:%S GMT", value)
    return value


def parse_date(ims):
    """ Parse rfc1123, rfc850 and asctime timestamps and return UTC epoch.
        (c)2014, Marcel Hellkamp
    """
    try:
        ts = email.utils.parsedate_tz(ims)
        return time.mktime(ts[:8] + (0, )) - (ts[9] or 0) - time.timezone
    except (TypeError, ValueError, IndexError, OverflowError):
        return None


def parse_auth(header):
    """ Parse rfc2617 HTTP authentication header string (basic) and return
        (user,pass) tuple or None
        (c)2014, Marcel Hellkamp
    """
    try:
        method, data = header.split(None, 1)
        if method.lower() == 'basic':
            data = base64.b64decode(uniorbytes(data, bytes))
            user, pwd = uniorbytes(data).split(':', 1)
            return user, pwd
    except (KeyError, AttributeError, ValueError):
        return (None, None)


def parse_range_header(header, maxlen=0):
    """ Yield (start, end) ranges parsed from a HTTP Range header. Skip
        unsatisfiable ranges. The end index is non-inclusive.
        (c)2014, Marcel Hellkamp
    """
    if not header or header[:6] != 'bytes=':
        return
    ranges = [r.split('-', 1) for r in header[6:].split(',') if '-' in r]
    for start, end in ranges:
        try:
            if not start:  # bytes=-100    -> last 100 bytes
                start, end = max(0, maxlen - int(end)), maxlen
            elif not end:  # bytes=100-    -> all but the first 99 bytes
                start, end = int(start), maxlen
            else:  # bytes=100-200 -> bytes 100-200 (inclusive)
                start, end = int(start), min(int(end) + 1, maxlen)
            if 0 <= start < end <= maxlen:
                yield start, end
        except ValueError:
            pass



# -*- coding: utf-8 -*-
"""

   `Vital String Tools`
--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--
    The MIT License (MIT) (c) 2016 Jared Lunde

"""
import re


camel_pattern = r"""([a-z])([A-Z])"""
camel_re = re.compile(camel_pattern)
def camel_to_underscore(string):
    """ Converts a CamelCased @string to camel_cased

        @string: #str object

        ..
            from vital.tools import camel_to_underscore

            camel_to_underscore("TedKoppel")
            # -> ted_koppel
        ..
    """
    return camel_re.sub(r"""\1_\2""", string).lower()


def underscore_to_camel(string):
    """ Converts an undescored_name @string to UnderscoredName

        @string: #str object

        ..
            from vital.tools import underscore_to_camel

            underscore_to_camel("ted_koppel")
            # -> TedKoppel
        ..
    """
    return "".join(s.capitalize() for s in string.split("_"))


non_alnum_re = re.compile(r"""([^A-Za-z0-9]+)""")
def to_alnum(string):
    """ Removes non-alphanumeric |[A-Za-z0-9]| characters from @string

        -> #str with only alphanumeric |[A-Za-z0-9]| characters

        ..
            import vital.tools
            print(vital.tools.to_alnum("2352@#%@#)gjgaoger"))
            # -> 2352gjgaoger
        ..
    """
    return non_alnum_re.sub("", string)


username_pattern = r"""([A-Za-z0-9_]+$)"""
username_re = re.compile(username_pattern)
def is_username(string, minlen=1, maxlen=15):
    """ Determines whether the @string pattern is username-like
        @string: #str being tested
        @minlen: minimum required username length
        @maxlen: maximum username length

        -> #bool
    """
    if string:
        string = string.strip()
        return username_re.match(string) and (minlen <= len(string) <= maxlen)
    return False


email_pattern = \
    r"""[A-Za-z0-9\.\+_-]+@[A-Za-z0-9_-]{1,63}(\.[a-zA-Z\-0-9]{0,24})*$"""
email_re = re.compile(email_pattern)
def is_email(string):
    """ Determines whether the @string pattern is email-like
        @string: #str being tested

        -> #bool
    """
    return True if email_re.match(string) else False


#
## ``JSON bigint to string functions for Javascript``
#
_NUMBERS = (int, float)
def bigint_to_string(val):
    """ Converts @val to a string if it is a big integer (|>2**53-1|)

        @val: #int or #float

        -> #str if @val is a big integer, otherwise @val
    """
    if isinstance(val, _NUMBERS) and not abs(val) <= 2**53-1:
        return str(val)
    return val


#: Recursive bigint to string parser
def rbigint_to_string(obj):
    """ Recursively converts big integers (|>2**53-1|) to strings

        @obj: Any python object

        -> @obj, with any big integers converted to #str objects
    """
    if isinstance(obj, (str, bytes)) or not obj:
        # the input is the desired one, return as is
        return obj
    elif hasattr(obj, 'items'):
        # the input is a dict {}
        for k, item in obj.items():
            obj[k] = rbigint_to_string(item)
        return obj
    elif hasattr(obj, '__iter__'):
        # the input is iterable
        is_tuple = isinstance(obj, tuple)
        if is_tuple:
            obj = list(obj)
        for i, item in enumerate(obj):
            obj[i] = rbigint_to_string(item)
        return obj if not is_tuple else tuple(obj)
    return bigint_to_string(obj)


#: Regex bigint to string (should be applied to valid json string)
bigint_re = re.compile(r"""(\:\s*)(9[0-9]{15,}|[0-9]{17,})([\,\]\}]*)""")
string_bigint_re = \
    re.compile(r"""(\:\s*)(")(9[0-9]{15,}|[0-9]{17,})(")([\s\,\]\}]*)""")
bigint_sub = bigint_re.sub
string_bigint_sub = bigint_re.sub
def json_bigint_to_string(s):
    """ Converts big integer-like values in JSON objects to strings via
        regex pattern. Any integer >= 9000000000000000 will be  replaced
        with a string.

        @s: a JSON encoded string

        -> @s with big integers replaced by string_bigint_sub

        ..
            from vital.tools import json_big_int_to_string
            print(json_big_int_to_string("{k: 9000000000000000}"))
            # -> {k: "9000000000000000"}
        ..
    """
    return bigint_sub(r"""\1"\2"\3""", s)


def json_bigint_from_string(s):
    """ Converts big integer-like string values in JSON objects to integers
        via regex pattern. Any integer-like string >= 9000000000000000 will be
        replaced with an integer.

        @s: a JSON encoded string

        -> @s with big integers replaced by string_bigint_sub

        ..
            from vital.tools import json_big_int_from_string
            print(json_big_int_from_string('{k: "9000000000000000"}'))
            # -> {k: 9000000000000000}
        ..
    """
    return string_bigint_sub(r"""\1"\3"\4""", s)


#: Username/Hashtag are in group \2
mentions_pattern = r"""(?:^|[^@\w])@(\w{1,15})([\b\s]|$)"""
mentions_re = re.compile(mentions_pattern)
try:
    # Wide UCS-4 build
    hashtag_pattern = (
        u"([\w'&%\$" +
        u'\U0001F300-\U0001F64F' +
        u'\U0001F680-\U0001F6FF' +
        u'\u2600-\u26FF\u2700-\u27BF' +
        u']+)')
    hashtag_re_pattern = \
        u'(?:^|[^#\w\@]*)#' + hashtag_pattern + u'([\b\s]|$)?'
    hashtag_re = re.compile(hashtag_re_pattern, re.UNICODE | re.M)
except re.error:
    # Narrow UCS-2 build
    hashtag_pattern = (
        u"([\w'&%\$" +
        u'\U0001F300-\U0001F64F' +
        u'\U0001F680-\U0001F6FF' +
        u'\u2600-\u26FF\u2700-\u27BF' +
        u']+)')
    hashtag_re_pattern = \
        u'(?:^|[^#\w\@]*)#' + hashtag_pattern + u'([\b\s]|$)?'
    hashtag_re = re.compile(hashtag_re_pattern, re.UNICODE | re.M)


def get_hashtags(s):
    """ Gets all #hashtag-like matches in a string

        -> #list of lowercase hashtag #str objects
    """
    return [r[0].lower() for r in hashtag_re.findall(str(s))]


def get_mentions(s):
    """ Gets all @mentions-like matches in a string
        -> #list of lowercase mention #str objects
    """
    return [r[0].lower() for r in mentions_re.findall(str(s))]


def remove_blank_lines(string):
    """ Removes all blank lines in @string

        -> #str without blank lines
    """
    return "\n".join(line
                     for line in string.split("\n")
                     if len(line.strip()))


def to_plural(num, singular, plural):
    if num == 1:
        return '1 %s' % singular
    return '%s %s' % (num, plural)



# -*- coding: utf-8 -*-
"""

   `Vital Encoding tools`
--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--
    The MIT License (MIT) (c) 2016 Jared Lunde

    Additional credits:
    http://blog.luminoso.com/2012/08/20/fix-unicode-mistakes-with-python/

"""
import sys
import decimal
import unicodedata


__all__ = (
    "uniorbytes",
    "recode_unicode",
    "fix_bad_unicode",
    "reinterpret_latin1_as_utf8",
    "fix_windows_1252_char",
    "reinterpret_windows1252_as_utf8",
    "reinterpret_latin1_as_windows1252",
    "text_badness",
    "text_cost",
    "WINDOWS_1252_GREMLINS",
    "WINDOWS_1252_CODEPOINTS",
    "SINGLE_BYTE_WEIRDNESS",
    "SINGLE_BYTE_LETTERS",
    "SCRIPT_TABLE",
    "stdout_encode"
)


def uniorbytes(s, result=str, enc="utf-8", err="strict"):
    """
    This function was made to avoid byte / str type errors received in
    packages like base64. Accepts all input types and will recursively
    encode entire lists and dicts.

    @s: the #bytes or #str item you are attempting to encode or decode
    @result: the desired output, either #str or #bytes
    @enc: the desired encoding
    @err: passed to :meth:bytes.decode, tells the decoder what to do about
        errors, e.g. 'replace'

    -> type specified in @result
    """
    if isinstance(s, result):
        # the input is the desired one, return as is
        return s

    if isinstance(s, (bytes, str)):
        # the input is either a byte or a string, convert to desired
        # result (result=bytes or str)
        if isinstance(s, bytes) and result == str:
            return s.decode(enc, err)
        elif isinstance(s, str) and result == bytes:
            return s.encode(enc)
        else:
            return str(s or ("" if s is None else s), enc)
    elif isinstance(s, (float, int, decimal.Decimal)):
        return uniorbytes(str(s), result, enc, err)
    elif isinstance(s, dict):
        # the input is a dict {}
        for k, item in list(s.items()):
            s[k] = uniorbytes(item, result=result, enc=enc, err=err)
        return s
    elif hasattr(s, '__iter__'):
        # the input is iterable
        for i, item in enumerate(s):
            s[i] = uniorbytes(item, result=result, enc=enc, err=err)
        return s
    return s


def recode_unicode(s, encoding='utf-8'):
    """ Inputs are encoded to utf-8 and then decoded to the desired
        output encoding

        @encoding: the desired encoding

        -> #str with the desired @encoding
    """
    if isinstance(s, str):
        return s.encode().decode(encoding)
    return s


def fix_bad_unicode(text):
    u"""Copyright:
        http://blog.luminoso.com/2012/08/20/fix-unicode-mistakes-with-python/

    Something you will find all over the place, in real-world text, is text
    that's mistakenly encoded as utf-8, decoded in some ugly format like
    latin-1 or even Windows codepage 1252, and encoded as utf-8 again.

    This causes your perfectly good Unicode-aware code to end up with garbage
    text because someone else (or maybe "someone else") made a mistake.

    This function looks for the evidence of that having happened and fixes it.
    It determines whether it should replace nonsense sequences of single-byte
    characters that were really meant to be UTF-8 characters, and if so, turns
    them into the correctly-encoded Unicode character that they were meant to
    represent.

    The input to the function must be Unicode. It's not going to try to
    auto-decode bytes for you -- then it would just create the problems it's
    supposed to fix.

        >>> print fix_bad_unicode(u'Ãºnico')
        único

        >>> print fix_bad_unicode(u'This text is fine already :þ')
        This text is fine already :þ

    Because these characters often come from Microsoft products, we allow
    for the possibility that we get not just Unicode characters 128-255, but
    also Windows's conflicting idea of what characters 128-160 are.

        >>> print fix_bad_unicode(u'This â€” should be an em dash')
        This — should be an em dash

    We might have to deal with both Windows characters and raw control
    characters at the same time, especially when dealing with characters like
    \x81 that have no mapping in Windows.

        >>> print fix_bad_unicode(u'This text is sad .â\x81”.')
        This text is sad .⁔.

    This function even fixes multiple levels of badness:

        >>> wtf = u'\xc3\xa0\xc2\xb2\xc2\xa0_\xc3\xa0\xc2\xb2\xc2\xa0'
        >>> print fix_bad_unicode(wtf)
        ಠ_ಠ

    However, it has safeguards against fixing sequences of letters and
    punctuation that can occur in valid text:

        >>> print fix_bad_unicode(u'not such a fan of Charlotte Brontë…”')
        not such a fan of Charlotte Brontë…”

    Cases of genuine ambiguity can sometimes be addressed by finding other
    characters that are not double-encoding, and expecting the encoding to
    be consistent:

        >>> print fix_bad_unicode(u'AHÅ™, the new sofa from IKEA®')
        AHÅ™, the new sofa from IKEA®

    Finally, we handle the case where the text is in a single-byte encoding
    that was intended as Windows-1252 all along but read as Latin-1:

        >>> print fix_bad_unicode(u'This text was never Unicode at all\x85')
        This text was never Unicode at all…
    """
    if not isinstance(text, str):
        return text

    if len(text) == 0:
        return text

    maxord = max(ord(char) for char in text)
    tried_fixing = []
    if maxord < 128:
        # Hooray! It's ASCII!
        return text
    else:
        attempts = [(text, text_badness(text) + len(text))]
        if maxord < 256:
            tried_fixing = reinterpret_latin1_as_utf8(text)
            tried_fixing2 = reinterpret_latin1_as_windows1252(text)
            attempts.append((tried_fixing, text_cost(tried_fixing)))
            attempts.append((tried_fixing2, text_cost(tried_fixing2)))
        elif all(ord(char) in WINDOWS_1252_CODEPOINTS for char in text):
            tried_fixing = reinterpret_windows1252_as_utf8(text)
            attempts.append((tried_fixing, text_cost(tried_fixing)))
        else:
            # We can't imagine how this would be anything but valid text.
            return text

        # Sort the results by badness
        attempts.sort(key=lambda x: x[1])
        goodtext = attempts[0][0]
        if goodtext == text:
            return goodtext
        else:
            return fix_bad_unicode(goodtext)


def reinterpret_latin1_as_utf8(wrongtext):
    """ :see:recode_unicode """
    newbytes = wrongtext.encode('latin-1', 'replace')
    return newbytes.decode('utf-8', 'replace')


def fix_windows_1252_char(char):
    if ord(char) in WINDOWS_1252_GREMLINS:
        return char.encode('WINDOWS_1252')
    return char.encode('latin-1', 'replace')


def reinterpret_windows1252_as_utf8(wrongtext):
    altered_bytes = [ fix_windows_1252_char(char) for char in wrongtext ]
    return b''.join(altered_bytes).decode('utf-8', 'replace')


def reinterpret_latin1_as_windows1252(wrongtext):
    """
    Maybe this was always meant to be in a single-byte encoding, and it
    makes the most sense in Windows-1252.
    """
    return wrongtext.encode('latin-1').decode('WINDOWS_1252', 'replace')


def text_badness(text):
    u'''
    Look for red flags that text is encoded incorrectly:

    Obvious problems:
    - The replacement character \ufffd, indicating a decoding error
    - Unassigned or private-use Unicode characters

    Very weird things:
    - Adjacent letters from two different scripts
    - Letters in scripts that are very rarely used on computers (and
      therefore, someone who is using them will probably get Unicode right)
    - Improbable control characters, such as 0x81

    Moderately weird things:
    - Improbable single-byte characters, such as ƒ or ¬
    - Letters in somewhat rare scripts
    '''
    assert isinstance(text, str)
    errors = 0
    very_weird_things = 0
    weird_things = 0
    prev_letter_script = None
    unicodedata_name = unicodedata.name
    unicodedata_category = unicodedata.category
    for char in text:
        index = ord(char)
        if index < 256:
            # Deal quickly with the first 256 characters.
            weird_things += SINGLE_BYTE_WEIRDNESS[index]
            if SINGLE_BYTE_LETTERS[index]:
                prev_letter_script = 'latin'
            else:
                prev_letter_script = None
        else:
            category = unicodedata_category(char)
            if category == 'Co':
                # Unassigned or private use
                errors += 1
            elif index == 0xfffd:
                # Replacement character
                errors += 1
            elif index in WINDOWS_1252_GREMLINS:
                lowchar = char.encode('WINDOWS_1252').decode('latin-1')
                weird_things += SINGLE_BYTE_WEIRDNESS[ord(lowchar)] - 0.5

            if category[0] == 'L':
                # It's a letter. What kind of letter? This is typically found
                # in the first word of the letter's Unicode name.
                name = unicodedata_name(char)
                scriptname = name.split()[0]
                freq, script = SCRIPT_TABLE.get(scriptname, (0, 'other'))
                if prev_letter_script:
                    if script != prev_letter_script:
                        very_weird_things += 1
                    if freq == 1:
                        weird_things += 2
                    elif freq == 0:
                        very_weird_things += 1
                prev_letter_script = script
            else:
                prev_letter_script = None

    return 100 * errors + 10 * very_weird_things + weird_things


def text_cost(text):
    """
    Assign a cost function to the length plus weirdness of a text string.
    """
    return text_badness(text) + len(text)


#######################################################################
# The rest of this file is esoteric info about characters, scripts, and their
# frequencies.
#
# Start with an inventory of "gremlins", which are characters from all over
# Unicode that Windows has instead assigned to the control characters
# 0x80-0x9F. We might encounter them in their Unicode forms and have to figure
# out what they were originally.

WINDOWS_1252_GREMLINS = {
    # adapted from http://effbot.org/zone/unicode-gremlins.htm
    0x0152,  # LATIN CAPITAL LIGATURE OE
    0x0153,  # LATIN SMALL LIGATURE OE
    0x0160,  # LATIN CAPITAL LETTER S WITH CARON
    0x0161,  # LATIN SMALL LETTER S WITH CARON
    0x0178,  # LATIN CAPITAL LETTER Y WITH DIAERESIS
    0x017E,  # LATIN SMALL LETTER Z WITH CARON
    0x017D,  # LATIN CAPITAL LETTER Z WITH CARON
    0x0192,  # LATIN SMALL LETTER F WITH HOOK
    0x02C6,  # MODIFIER LETTER CIRCUMFLEX ACCENT
    0x02DC,  # SMALL TILDE
    0x2013,  # EN DASH
    0x2014,  # EM DASH
    0x201A,  # SINGLE LOW-9 QUOTATION MARK
    0x201C,  # LEFT DOUBLE QUOTATION MARK
    0x201D,  # RIGHT DOUBLE QUOTATION MARK
    0x201E,  # DOUBLE LOW-9 QUOTATION MARK
    0x2018,  # LEFT SINGLE QUOTATION MARK
    0x2019,  # RIGHT SINGLE QUOTATION MARK
    0x2020,  # DAGGER
    0x2021,  # DOUBLE DAGGER
    0x2022,  # BULLET
    0x2026,  # HORIZONTAL ELLIPSIS
    0x2030,  # PER MILLE SIGN
    0x2039,  # SINGLE LEFT-POINTING ANGLE QUOTATION MARK
    0x203A,  # SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
    0x20AC,  # EURO SIGN
    0x2122,  # TRADE MARK SIGN
}


# a list of Unicode characters that might appear in Windows-1252 text
WINDOWS_1252_CODEPOINTS = set(list(range(256)) +
                              list(WINDOWS_1252_GREMLINS))


# Rank the characters typically represented by a single byte -- that is, in
# Latin-1 or Windows-1252 -- by how weird it would be to see them in running
# text.
#
#   0 = not weird at all
#   1 = rare punctuation or rare letter that someone could certainly
#       have a good reason to use. All Windows-1252 gremlins are at least
#       weirdness 1.
#   2 = things that probably don't appear next to letters or other
#       symbols, such as math or currency symbols
#   3 = obscure symbols that nobody would go out of their way to use
#       (includes symbols that were replaced in ISO-8859-15)
#   4 = why would you use this?
#   5 = unprintable control character
#
# The Portuguese letter Ã (0xc3) is marked as weird because it would usually
# appear in the middle of a word in actual Portuguese, and meanwhile it
# appears in the mis-encodings of many common characters.


SINGLE_BYTE_WEIRDNESS = (
#   0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f
    5, 5, 5, 5, 5, 5, 5, 5, 5, 0, 0, 5, 5, 5, 5, 5,  # 0x00
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,  # 0x10
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  # 0x20
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  # 0x30
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  # 0x40
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  # 0x50
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  # 0x60
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5,  # 0x70
    2, 5, 1, 4, 1, 1, 3, 3, 4, 3, 1, 1, 1, 5, 1, 5,  # 0x80
    5, 1, 1, 1, 1, 3, 1, 1, 4, 1, 1, 1, 1, 5, 1, 1,  # 0x90
    1, 0, 2, 2, 3, 2, 4, 2, 4, 2, 2, 0, 3, 1, 1, 4,  # 0xa0
    2, 2, 3, 3, 4, 3, 3, 2, 4, 4, 4, 0, 3, 3, 3, 0,  # 0xb0
    0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  # 0xc0
    1, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0,  # 0xd0
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  # 0xe0
    1, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0,  # 0xf0
)


# Pre-cache the Unicode data saying which of these first 256 characters
# are letters. We'll need it often.
try:
    SINGLE_BYTE_LETTERS = [
        unicodedata.category(chr(i))[0] == 'L'
        for i in range(256)
    ]
except TypeError:
    SINGLE_BYTE_LETTERS = [
        unicodedata.category(unichr(i))[0] == 'L'
        for i in range(256)
    ]

# A table telling us how to interpret the first word of a letter's Unicode
# name. The number indicates how frequently we expect this script to be
# used  on computers. Many scripts not included here are assumed to have
# a frequency of "0" -- if you're going to write in Linear B using
# Unicode, you're probably aware enough of encoding issues to get it
# right.
#
# The lowercase name is a general category -- for example, Han characters
# and Hiragana characters are very frequently adjacent in Japanese, so
# they all go into category 'cjk'. Letters of different categories are
# assumed not to appear next to each other often.
SCRIPT_TABLE = {
    'LATIN': (3, 'latin'),
    'CJK': (2, 'cjk'),
    'ARABIC': (2, 'arabic'),
    'CYRILLIC': (2, 'cyrillic'),
    'GREEK': (2, 'greek'),
    'HEBREW': (2, 'hebrew'),
    'KATAKANA': (2, 'cjk'),
    'HIRAGANA': (2, 'cjk'),
    'HIRAGANA-KATAKANA': (2, 'cjk'),
    'HANGUL': (2, 'cjk'),
    'DEVANAGARI': (2, 'devanagari'),
    'THAI': (2, 'thai'),
    'FULLWIDTH': (2, 'cjk'),
    'MODIFIER': (2, None),
    'HALFWIDTH': (1, 'cjk'),
    'BENGALI': (1, 'bengali'),
    'LAO': (1, 'lao'),
    'KHMER': (1, 'khmer'),
    'TELUGU': (1, 'telugu'),
    'MALAYALAM': (1, 'malayalam'),
    'SINHALA': (1, 'sinhala'),
    'TAMIL': (1, 'tamil'),
    'GEORGIAN': (1, 'georgian'),
    'ARMENIAN': (1, 'armenian'),
    'KANNADA': (1, 'kannada'),  # mostly used for looks of disapproval
    'MASCULINE': (1, 'latin'),
    'FEMININE': (1, 'latin')
}


def stdout_encode(u, default='utf8'):
    """ Encodes a given string with the proper standard out encoding
        If sys.stdout.encoding isn't specified, it this defaults
        to @default

        @default: default encoding

        -> #str with standard out encoding
    """
    encoding = sys.stdout.encoding or default
    if encoding != 'utf8':
        return u.encode(encoding, "replace").decode(encoding, "replace")
    return u



# -*- coding: utf-8 -*-
"""

   `Vital Security` helpful functions for securing data
--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--
    The MIT License (MIT) (c) 2016 Jared Lunde

"""
import os
import random
import string
from math import ceil
from cmath import log, exp

import hmac
from Crypto.Cipher import AES
from hashlib import sha256

from base64 import b64encode, b64decode
try:
    import ujson as json
except ImportError:
    import json

from vital.tools.encoding import uniorbytes


#
#  ``General cryptography``
#
def aes_b64_encrypt(value, secret, block_size=AES.block_size):
    """ AES encrypt @value with @secret using the |CFB| mode of AES
        with a cryptographically secure initialization vector.

        -> (#str) AES encrypted @value

        ..
            from vital.security import aes_encrypt, aes_decrypt
            aes_encrypt("Hello, world",
                        "aLWEFlwgwlreWELFNWEFWLEgwklgbweLKWEBGW")
            # -> 'zYgVYMbeOuiHR50aMFinY9JsfyMQCvpzI+LNqNcmZhw='
            aes_decrypt(
                "zYgVYMbeOuiHR50aMFinY9JsfyMQCvpzI+LNqNcmZhw=",
                "aLWEFlwgwlreWELFNWEFWLEgwklgbweLKWEBGW")
            # -> 'Hello, world'
        ..
    """
    iv = randstr(block_size * 2, rng=random)
    cipher = AES.new(secret[:32], AES.MODE_CFB, iv[:block_size].encode())
    return iv + b64encode(cipher.encrypt(
        uniorbytes(value, bytes))).decode('utf-8')


def aes_b64_decrypt(value, secret, block_size=AES.block_size):
    """ AES decrypt @value with @secret using the |CFB| mode of AES
        with a cryptographically secure initialization vector.

        -> (#str) AES decrypted @value

        ..
            from vital.security import aes_encrypt, aes_decrypt
            aes_encrypt("Hello, world",
                        "aLWEFlwgwlreWELFNWEFWLEgwklgbweLKWEBGW")
            # -> 'zYgVYMbeOuiHR50aMFinY9JsfyMQCvpzI+LNqNcmZhw='
            aes_decrypt(
                "zYgVYMbeOuiHR50aMFinY9JsfyMQCvpzI+LNqNcmZhw=",
                "aLWEFlwgwlreWELFNWEFWLEgwklgbweLKWEBGW")
            # -> 'Hello, world'
        ..
    """
    if value is not None:
        iv = value[:block_size]
        cipher = AES.new(secret[:32], AES.MODE_CFB, iv)
        return cipher.decrypt(b64decode(
            uniorbytes(value[block_size * 2:], bytes))).decode('utf-8')


def aes_encrypt(value, secret, block_size=AES.block_size):
    """ AES encrypt @value with @secret using the |CFB| mode of AES
        with a cryptographically secure initialization vector.

        -> (#bytes) AES encrypted @value

        ..
            from vital.security import aes_encrypt, aes_decrypt
            aes_encrypt("Hello, world",
                        "aLWEFlwgwlreWELFNWEFWLEgwklgbweLKWEBGW")
            # -> 'zYgVYMbeOuiHR50aMFinY9JsfyMQCvpzI+LNqNcmZhw='
            aes_decrypt(
                "zYgVYMbeOuiHR50aMFinY9JsfyMQCvpzI+LNqNcmZhw=",
                "aLWEFlwgwlreWELFNWEFWLEgwklgbweLKWEBGW")
            # -> 'Hello, world'
        ..f
    """
    iv = os.urandom(block_size * 2)
    cipher = AES.new(secret[:32], AES.MODE_CFB, iv[:block_size])
    return b'%s%s' % (iv, cipher.encrypt(value))


def aes_decrypt(value, secret, block_size=AES.block_size):
    """ AES decrypt @value with @secret using the |CFB| mode of AES
        with a cryptographically secure initialization vector.

        -> (#str) AES decrypted @value

        ..
            from vital.security import aes_encrypt, aes_decrypt
            aes_encrypt("Hello, world",
                        "aLWEFlwgwlreWELFNWEFWLEgwklgbweLKWEBGW")
            # -> 'zYgVYMbeOuiHR50aMFinY9JsfyMQCvpzI+LNqNcmZhw='
            aes_decrypt(
                "zYgVYMbeOuiHR50aMFinY9JsfyMQCvpzI+LNqNcmZhw=",
                "aLWEFlwgwlreWELFNWEFWLEgwklgbweLKWEBGW")
            # -> 'Hello, world'
        ..
    """
    if value is not None:
        cipher = AES.new(secret[:32], AES.MODE_CFB, value[:block_size])
        return cipher.decrypt(uniorbytes(value[block_size * 2:], bytes))


def aes_pad(s, block_size=32, padding='{'):
    """ Adds padding to get the correct block sizes for AES encryption

        @s: #str being AES encrypted or decrypted
        @block_size: the AES block size
        @padding: character to pad with

        -> padded #str

        ..
            from vital.security import aes_pad
            aes_pad("swing")
            # -> 'swing{{{{{{{{{{{{{{{{{{{{{{{{{{{'
        ..
    """
    return s + (block_size - len(s) % block_size) * padding


def aes_unpad(s, block_size=32, padding='{'):
    """ Removes padding to get the value from @s for AES decryption

        @s: #str being AES encrypted or decrypted
        @block_size: the AES block size
        @padding: character to pad with

        -> unpadded #str

        ..
            from vital.security import aes_pad
            aes_unpad("swing{{{{{{{{{{{{{{{{{{{{{{{{{{{")
            # -> 'swing'
        ..
    """
    return s.rstrip(padding)


def lscmp(a, b):
    """ Compares two strings in a cryptographically safe way:
        Runtime is not affected by length of common prefix, so this
        is helpful against timing attacks.

        ..
            from vital.security import lscmp
            lscmp("ringo", "starr")
            # -> False
            lscmp("ringo", "ringo")
            # -> True
        ..
    """
    l = len
    return not sum(0 if x == y else 1 for x, y in zip(a, b)) and l(a) == l(b)


#
#  ``Cookies``
#
def cookie(data, key_salt='', secret=None, digestmod=None):
    """ Encodes or decodes a signed cookie.
        @data: cookie data
        @key_salt: HMAC key signing salt
        @secret: HMAC signing secret key
        @digestmod: hashing algorithm to sign with, recommended >=sha256

        -> HMAC signed or unsigned cookie data

        ..
            from vital.security import cookie

            cookie("Hello, world.", "saltyDog", secret="alBVlwe")
            # -> '!YuOoKwDp8GhrwwojdjTxSCj1c2Z+7yz7r6cC7E3hBWo=?IkhlbGxvLCB3b3JsZC4i'
            cookie(
                "!YuOoKwDp8GhrwwojdjTxSCj1c2Z+7yz7r6cC7E3hBWo=?IkhlbGxvLCB3b3JsZC4i",
                "saltyDog", secret="alBVlwe")
            # -> 'Hello, world.'
        ..
    """
    digestmod = digestmod or sha256
    if not data:
        return None
    try:
        # Decode signed cookie
        assert cookie_is_encoded(data)
        datab = uniorbytes(data, bytes)
        sig, msg = datab.split(uniorbytes('?', bytes), 1)
        key = ("{}{}").format(secret, key_salt)
        sig_check = hmac.new(
            key=uniorbytes(key, bytes), msg=msg, digestmod=digestmod).digest()
        sig_check = uniorbytes(b64encode(sig_check), bytes)
        if lscmp(sig[1:], sig_check):
            return json.loads(uniorbytes(b64decode(msg)))
        return None
    except:
        # Encode and sign a json-able object. Return a string.
        key = ("{}{}").format(secret, key_salt)
        msg = b64encode(uniorbytes(json.dumps(data), bytes))
        sig = hmac.new(
            key=uniorbytes(key, bytes), msg=msg,
            digestmod=digestmod).digest()
        sig = uniorbytes(b64encode(sig), bytes)
        return uniorbytes('!'.encode() + sig + '?'.encode() + msg)


def cookie_is_encoded(data):
    """ Tests whether or not a cookie is encoded / HMAC signed

        -> #bool True if encoded

        ..
            from vital.security import cookie_is_encoded

            cookie_is_encoded(
                "!YuOoKwDp8GhrwwojdjTxSCj1c2Z+7yz7r6cC7E3hBWo=?IkhlbGxvLCB3b3JsZC4i")
            # -> True
        ..
    """
    return data.startswith('!') and '?' in data


def strkey(val, chaffify=1, keyspace=string.ascii_letters+string.digits):
    """ Converts integers to a sequence of strings, and reverse.
        This is not intended to obfuscate numbers in any kind of
        cryptographically secure way, in fact it's the opposite. It's
        for predictable, reversable, obfuscation. It can also be used to
        transform a random bit integer to a string of the same bit
        length.

        @val: #int or #str
        @chaffify: #int multiple to avoid 1=b, 2=c, ... obfuscates the ordering
        @keyspace: #str allowed output chars

        -> #str if @val is #int, #int if @val is #str

        ..
            from vital.security import strkey

            strkey(0, chaffify=1)
            # -> b
            strkey(0, chaffify=4)
            # -> e
            strkey(90000000000050500502200302035023)
            # -> 'f3yMpJQUazIZHp1UO7k'
            strkey('f3yMpJQUazIZHp1UO7k')
            # -> 90000000000050500502200302035023
            strkey(2000000, chaffify=200000000000)
            # -> 'DIaqtyo2sC'
        ..
    """
    chaffify = chaffify or 1
    keylen = len(keyspace)
    try:
        # INT TO STRING
        # must be a positive integer
        if val < 2:
            raise ValueError("Input value must be greater than 1.")
        # chaffify the value
        val = val * chaffify
        # output the new string value
        out = []
        out_add = out.append
        while val > 0:
            val, digit = divmod(val, keylen)
            out_add(keyspace[digit])
        return "".join(out)[::-1]
    except TypeError:
        # STRING TO INT
        out = 0
        val = str(val)
        find = str.find
        for c in val:
            out = out * keylen + find(keyspace, c)
        # dechaffify the value
        out = out // chaffify
        return int(out)


#
# ``Random token generation``
#
def randhex(size=32, rng=None):
    """ Gets a random hex string of @size in terms of number of characters.
        This is an extremely fast way to generate a random string.

        @size: (#int) approximate size of the hex to generate in number
            of characters
        @rng: the random number generator to use, :class:random.SystemRandom
            is used by default.
    """
    return "%0x" % (rng or random.SystemRandom()).getrandbits(size * 4)


def calc_chars_in(bits, keyspace):
    if bits > 512:
        raise ValueError('Bits must be <= 512')
    bits = bits * log(2)
    return log(exp(bits))/log(keyspace)


def chars_in(bits, keyspace):
    """ ..
        log2(keyspace^x_chars) = bits
        log(keyspace^x_chars) = log(2) * bits
        exp(log(keyspace^x_chars)) = exp(log(2) * bits)
        x_chars = log(exp(log(2) * bits)) / log(keyspace)
        ..
        -> (#int) number of characters in @bits of entropy given the @keyspace
    """
    keyspace = len(keyspace)
    if keyspace < 2:
        raise ValueError("Keyspace size must be >1")
    bits_per_cycle = 512
    if bits > bits_per_cycle:
        chars = 0
        bits_processed = 0
        cycles = ceil(bits / bits_per_cycle)
        for _ in range(int(cycles)):
            if bits_processed + bits_per_cycle > bits:
                bits_per_cycle = bits - bits_processed
            chars += calc_chars_in(bits_per_cycle, keyspace)
            bits_processed += bits_per_cycle
    else:
        chars = calc_chars_in(bits, keyspace)
    return abs(chars)


def calc_bits_in(length, keyspace):
    if length > 64:
        raise ValueError('Length must be <= 64')
    return log(keyspace**length, 2)


def bits_in(length, keyspace):
    """ |log2(keyspace^length) = bits|
        -> (#float) number of bits of entropy in @length of characters for
            a given a @keyspace
    """
    keyspace = len(keyspace)
    length_per_cycle = 64
    if length > length_per_cycle:
        bits = 0
        length_processed = 0
        cycles = ceil(length / length_per_cycle)
        for _ in range(int(cycles)):
            if length_processed + length_per_cycle > length:
                length_per_cycle = length - length_processed
            bits += calc_bits_in(length_per_cycle, keyspace)
            length_processed += length_per_cycle
    else:
        bits = calc_bits_in(length, keyspace)
    return float(abs(bits))


def iter_random_chars(bits,
                      keyspace=string.ascii_letters + string.digits + '#/.',
                      rng=None):
    """ Yields a cryptographically secure random key of desired @bits of
        entropy within @keyspace using :class:random.SystemRandom

        @bits: (#int) minimum bits of entropy
        @keyspace: (#str) or iterable allowed output chars

        ..
            from vital.security import iter_rand

            for char in iter_rand(512):
                do_something_with(char)
    """
    if bits < 8:
        raise ValueError('Bits cannot be <8')
    else:
        chars = chars_in(bits, keyspace)
    rng = rng or random.SystemRandom()
    for char in range(int(ceil(chars))):
        yield rng.choice(keyspace)


def randkey(bits, keyspace=string.ascii_letters + string.digits + '#/.',
            rng=None):
    """ Returns a cryptographically secure random key of desired @bits of
        entropy within @keyspace using :class:random.SystemRandom

        @bits: (#int) minimum bits of entropy
        @keyspace: (#str) or iterable allowed output chars
        @rng: the random number generator to use. Defaults to
            :class:random.SystemRandom. Must have a |choice| method

        -> (#str) random key

        ..
            from vital.security import randkey

            randkey(24)
            # -> '9qaX'
            randkey(48)
            # -> 'iPJ5YWs9'
            randkey(64)
            # - > 'C..VJ.KLdxg'
            randkey(64, keyspace="abc", rng=random)
            # -> 'aabcccbabcaacaccccabcaabbabcacabacbbbaaab'
        ..
    """
    return "".join(char for char in iter_random_chars(bits, keyspace, rng))


def randstr(size, keyspace=string.ascii_letters + string.digits, rng=None):
    """ Returns a cryptographically secure random string of desired @size
        (in character length) within @keyspace using :class:random.SystemRandom

        @size: (#int) number of random characters to generate
        @keyspace: (#str) or iterable allowed output chars
        @rng: the random number generator to use. Defaults to
            :class:random.SystemRandom. Must have a |choice| method

        -> #str random key

        ..
            from vital.security import randkey

            randstr(4)
            # -> '9qaX'
        ..
    """
    rng = rng or random.SystemRandom()
    return "".join(rng.choice(keyspace) for char in range(int(ceil(size))))



import sys


def compat(min_version):
    ''' @min_version: (#str) minimum version number formatted like |2.7|
            or |2.7.6|
        -> (#bool) |True| if the system version is at least @min_version
    '''
    return sys.version_info >= tuple(map(int, min_version.split('.')))



# -*- coding: utf-8 -*-
"""

   `Vital List Tools`
--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--
    The MIT License (MIT) (c) 2016 Jared Lunde

"""
from random import SystemRandom as rng

try:
    from collections import UserList
except ImportError:
    UserList = list


def unique_list(seq):
    """ Removes duplicate elements from given @seq

        @seq: a #list or sequence-like object

        -> #list
    """
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def pairwise(seq):
    """ @seq: (#iterable)

        -> yields pairs
        ..
            for pair in pairwise([0, 1, 2, 3, 4, 5]):
                print(pair)
            # (0, 1)
            # (2, 3)
            # (4, 5)
        ..
    """
    return grouped(seq, 2)


def grouped(seq, size):
    """ @seq: (#iterable)

        -> yields groups
        ..
            lst = [0, 1, 2, 3, 4, 5]
            for group in grouped(lst, 3):
                print(group)
            # (0, 1, 2)
            # (3, 4, 5)
        ..
    """
    return zip(*[iter(seq)] * size)


def flatten(*seqs):
    """ Flattens a sequence e.g. |[(1, 2), (3, (4, 5))] -> [1, 2, 3, 4, 5]|

        @seq: #tuple, #list or :class:UserList

        -> yields an iterator

        ..
            l = [(1, 2), (3, 4)]
            for x in flatten(l):
                print(x)
        ..
    """
    for seq in seqs:
        for item in seq:
            if isinstance(item, (tuple, list, UserList)):
                for subitem in flatten(item):
                    yield subitem
            else:
                yield item


def remove_empty(seq):
    """ Removes #None types from sequences

        @seq: a #list or sequence-like object

        -> #list
    """
    return [x for x in seq if x is not None]


def randrange(seq):
    """ Yields random values from @seq until @seq is empty """
    seq = seq.copy()
    choose = rng().choice
    remove = seq.remove
    for x in range(len(seq)):
        y = choose(seq)
        remove(y)
        yield y



# -*- coding: utf-8 -*-
"""

   `Vital local caching tools`
--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--
    The MIT License (MIT) (c) 2016 Jared Lunde

"""
import sys
from threading import local

from vital.cache.decorators import *

_all = [
  'local_lru',
  'typed_lru',
  'local_expiring_lru',
  'cached_property',
  'DictProperty',
  'memoize',
  'pickle_memoize',
  'sweet_pickle',
  'high_pickle'
  'local_property'
]

from vital.tools import systools

if systools.compat('3.4'):
    from vital.cache.async_decorators import async_lru
    _all.append(async_lru)


__all__ = _all


def local_property():
    """ Property structure which maps within the :func:local() thread
        (c)2014, Marcel Hellkamp
    """
    ls = local()

    def fget(self):
        try:
            return ls.var
        except AttributeError:
            raise RuntimeError("Request context not initialized.")

    def fset(self, value):
        ls.var = value

    def fdel(self):
        del ls.var

    return property(fget, fset, fdel, 'Thread-local property')



# -*- coding: utf-8 -*-
"""

   `Vital Dict Tools`
--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--
    The MIT License (MIT) (c) 2016 Jared Lunde

"""
from itertools import chain
from collections import OrderedDict


__all__ = (
    "merge_dict",
    "revrank_dict",
    "rank_dict",
    "getitem_in")


def merge_dict(*dicts):
    """ Merges given @*dicts into one #dict and returns a new dict
        without changing the state of the inputs.

        @*dicts: one or several #dict objects

        -> new merged #dict
    """
    return dict(chain(*list(d.items() for d in dicts)))


def revrank_dict(dict, key=lambda t: t[1], as_tuple=False):
    """ Reverse sorts a #dict by a given key, optionally returning it as a
        #tuple. By default, the @dict is sorted by it's value.

        @dict: the #dict you wish to sorts
        @key: the #sorted key to use
        @as_tuple: returns result as a #tuple ((k, v),...)

        -> :class:OrderedDict or #tuple
    """
    sorted_list = sorted(dict.items(), key=key, reverse=True)
    return OrderedDict(sorted_list) if not as_tuple else tuple(sorted_list)


def rank_dict(dict, key=lambda t: t[1], as_tuple=False):
    """ Sorts a #dict by a given key, optionally returning it as a
        #tuple. By default, the @dict is sorted by it's value.

        @dict: the #dict you wish to sorts
        @key: the #sorted key to use
        @as_tuple: returns result as a #tuple ((k, v),...)

        -> :class:OrderedDict or #tuple
    """
    sorted_list = sorted(dict.items(), key=key)
    return OrderedDict(sorted_list) if not as_tuple else tuple(sorted_list)


def getitem_in(obj, name):
    """ Finds a key in @obj via a period-delimited string @name.
        @obj: (#dict)
        @name: (#str) |.|-separated keys to search @obj in
        ..
            obj = {'foo': {'bar': {'baz': True}}}
            getitem_in(obj, 'foo.bar.baz')
        ..
        |True|
    """
    for part in name.split('.'):
        obj = obj[part]
    return obj



# -*- coding: utf-8 -*-
"""

  `Vital HTML Tools`
--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--
   The MIT License (MIT) (c) 2016 Jared Lunde

"""
import re
import uuid

try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote

from vital.tools.strings import hashtag_re, mentions_re


__all__ = (
    "whitespace_re",
    "js_comments_re",
    "remove_whitespace",
    "hashtag_links",
    "mentions_links",
    "nl2p",
    "escape_json"
)


whitespace_re = re.compile(r"""\s+""").sub
js_comments_re = re.compile(
    r"""(?:\/\*(?:[\s\S]*?)\*\/)|(?:([\s])+\/\/(?:.*)$)""", re.M).sub
html_ignore_whitespace_re = re.compile(
    r"""<(script|textarea|code|pre)>(.*?(?!<\1>).+?)</\s*\1>""", re.DOTALL)


def remove_whitespace(s):
    """ Unsafely attempts to remove HTML whitespace. This is not an HTML parser
        which is why its considered 'unsafe', but it should work for most
        implementations. Just use on at your own risk.

        @s: #str

        -> HTML with whitespace removed, ignoring <pre>, script, textarea and code
            tags
    """
    ignores = {}
    for ignore in html_ignore_whitespace_re.finditer(s):
        name = "{}{}{}".format(r"{}", uuid.uuid4(), r"{}")
        ignores[name] = ignore.group()
        s = s.replace(ignore.group(), name)
    s = whitespace_re(r' ', s).strip()
    for name, val in ignores.items():
        s = s.replace(name, val)
    return s


def hashtag_links(uri, s):
    """ Turns hashtag-like strings into HTML links

        @uri: /uri/ root for the hashtag-like
        @s: the #str string you're looking for |#|hashtags in

        -> #str HTML link |<a href="/uri/hashtag">hashtag</a>|
    """
    for tag, after in hashtag_re.findall(s):
        _uri = '/' + (uri or "").lstrip("/") + quote(tag)
        link = '<a href="{}">#{}</a>{}'.format(_uri.lower(), tag, after)
        s = s.replace('#' + tag, link)
    return s


def mentions_links(uri, s):
    """ Turns mentions-like strings into HTML links,
        @uri: /uri/ root for the hashtag-like
        @s: the #str string you're looking for |@|mentions in

        -> #str HTML link |<a href="/uri/mention">mention</a>|
    """
    for username, after in mentions_re.findall(s):
        _uri = '/' + (uri or "").lstrip("/") + quote(username)
        link = '<a href="{}">@{}</a>{}'.format(_uri.lower(), username, after)
        s = s.replace('@' + username, link)
    return s


_p_re = re.compile(r'(?:\r\n|\r|\n){2,}')


def nl2p(string):
    return ''.join(('<p>%s</p>' % s.replace('\n', '<br>')
                   for s in _p_re.split(string)))


def escape_json(string):
    return string.replace('\\', '\\\\')\
                 .replace('"', '\\"')



#!/usr/bin/python3 -S
import os
import sys
import uuid
from setuptools import setup
from pip.req import parse_requirements
from pkgutil import walk_packages


def compat(min_version):
    ''' @min_version: (#str) minimum version number formatted like |2.7|
            or |2.7.6|
        -> (#bool) |True| if the system version is at least @min_version
    '''
    return sys.version_info >= tuple(map(int, min_version.split('.')))


PKG = 'vital'
PKG_NAME = 'vital-tools'
PKG_VERSION = '0.1.8'

pathname = os.path.dirname(os.path.realpath(__file__))


# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_reqs = parse_requirements(pathname + "/requirements.txt",
                                  session=uuid.uuid1())


def find_packages(prefix=""):
    path = [prefix]
    yield prefix
    prefix = prefix + "."
    for _, name, ispkg in walk_packages(path, prefix):
        if ispkg:
            if name == 'vital.cache.async_decorators' and not compat('3.4'):
                pass
            else:
                yield name


setup(
    name=PKG_NAME,
    version=PKG_VERSION,
    description='Vital tools for writing Python 3.4+ packages.',
    author='Jared Lunde',
    author_email='jared.lunde@gmail.com',
    url='https://github.com/jaredlunde/vital-tools',
    license="MIT",
    install_requires=[str(ir.req) for ir in install_reqs],
    packages=list(find_packages(PKG))
)



# -*- coding: utf-8 -*-
"""

   `Vital`
    Essential tools for building Python 3 applications.
--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--·--
    The MIT License (MIT) (c) 2016 Jared Lunde
    http://github.com/jaredlunde/vital-tools

"""
from vital import cache
from vital import security
from vital import tools
if tools.systools.compat('3.3'):
    from vital import debug


__author__ = "Jared Lunde"
__license__ = "MIT"
__version__ = "0.1.8"
